from typing import List
import asyncio
from dataclasses import dataclass
from busline.event.event import Event
from busline.client.subscriber.event_handler.event_handler import EventHandler


@dataclass
class MultiEventHandler(EventHandler):
    """
    Call a batch of pre-defined handlers.

    Another parameter can be specified. If strict order is False, async capabilities can be exploited

    Author: Nicola Ricciardi
    """

    handlers: List[EventHandler]
    strict_order: bool = False


    async def handle(self, topic: str, event: Event):
        if self.strict_order:
            for handler in self.handlers:
                await handler.handle(topic, event)
        else:
            tasks = [handler.handle(topic, event) for handler in self.handlers]

            await asyncio.gather(*tasks)
            


from abc import ABC, abstractmethod
from typing import Dict

from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.event.event import Event


class SchemafullEventHandler(EventHandler, ABC):

    @abstractmethod
    def input_schema(self) -> Dict:
        raise NotImplemented()

    @abstractmethod
    def output_schema(self) -> Dict:
        raise NotImplemented()


import logging
import asyncio
from dataclasses import dataclass
from busline.event.event import Event
from busline.local.eventbus.eventbus import EventBus


@dataclass
class AsyncLocalEventBus(EventBus):
    """
    Async local eventbus

    Author: Nicola Ricciardi
    """

    async def put_event(self, topic: str, event: Event):

        topic_subscriptions = self._get_topic_subscriptions(topic)

        logging.debug(f"new event {event} on topic {topic}, notify subscribers: {topic_subscriptions}")

        tasks = [subscriber.on_event(topic, event) for subscriber in topic_subscriptions]

        await asyncio.gather(*tasks)

            



import asyncio
from dataclasses import dataclass, field
from typing import Optional, override, List, Self

from busline.client.client import EventBusClient
from busline.client.publisher.publisher import Publisher
from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.client.subscriber.topic_subscriber import TopicSubscriber
from busline.event.event import Event
from busline.client.subscriber.subscriber import Subscriber


@dataclass
class PubSubClient(EventBusClient):
    """
    Eventbus client which should used by components which wouldn't be a publisher/subscriber, but they need them

    Author: Nicola Ricciardi
    """

    publishers: List[Publisher]
    subscribers: List[Subscriber]

    @classmethod
    def from_pubsub(cls, publisher: Optional[Publisher] = None, subscriber: Optional[Subscriber] = None) -> Self:

        publishers = []
        if publisher is not None:
            publishers = [publisher]

        subscribers = []
        if subscriber is not None:
            subscribers = [subscriber]

        return cls(publishers, subscribers)

    @classmethod
    def from_pubsub_client(cls, client: Self) -> Self:
        return cls(client.publishers.copy(), client.subscribers.copy())

    @override
    async def connect(self):
        """
        Connect all publishers and subscribers
        """

        tasks = [publisher.connect() for publisher in self.publishers]
        tasks += [subscriber.connect() for subscriber in self.subscribers]

        await asyncio.gather(*tasks)

    @override
    async def disconnect(self):
        """
        Disconnect all publishers and subscribers
        """

        tasks = [publisher.disconnect() for publisher in self.publishers]
        tasks += [subscriber.disconnect() for subscriber in self.subscribers]

        await asyncio.gather(*tasks)

    @override
    async def publish(self, topic: str, event: Event, **kwargs):
        """
        Publish event using all publishers
        """

        await asyncio.gather(*[
            publisher.publish(topic, event, **kwargs) for publisher in self.publishers
        ])

    @override
    async def subscribe(self, topic: str, **kwargs):
        """
        Subscribe all subscribers on topic
        """

        await asyncio.gather(*[
            subscriber.subscribe(topic, **kwargs) for subscriber in self.subscribers
        ])

    @override
    async def unsubscribe(self, topic: Optional[str] = None, **kwargs):
        """
        Alias of `client.subscriber.unsubscribe(...)`
        """

        await asyncio.gather(*[
            subscriber.unsubscribe(topic, **kwargs) for subscriber in self.subscribers
        ])


@dataclass
class PubTopicSubClient(PubSubClient):
    """
    Eventbus client which should used by components which wouldn't be a publisher/subscriber, but they need them

    Author: Nicola Ricciardi
    """

    subscribers: List[TopicSubscriber]

    @override
    async def subscribe(self, topic: str, handler: Optional[EventHandler] = None, **kwargs):
        """
        Subscribe all subscribers on topic
        """

        await asyncio.gather(*[
            subscriber.subscribe(topic, handler=handler, **kwargs) for subscriber in self.subscribers
        ])


@dataclass
class PubSubClientBuilder:
    """
    Builder for a pub/sub client.

    Author: Nicola Ricciardi
    """

    base_client: PubSubClient = field(
        default_factory=lambda: PubSubClient([], []),
        kw_only=True
    )


    def with_publisher(self, publisher: Publisher) -> Self:
        self.base_client.publishers.append(publisher)

        return self

    def with_publishers(self, publishers: List[Publisher]) -> Self:
        self.base_client.publishers.extend(publishers)

        return self

    def with_subscriber(self, subscriber: Subscriber) -> Self:
        self.base_client.subscribers.append(subscriber)

        return self

    def with_subscribers(self, subscribers: List[Subscriber]) -> Self:
        self.base_client.subscribers.extend(subscribers)

        return self

    def build(self) -> PubSubClient:
        return self.base_client


@dataclass
class PubTopicSubClientBuilder(PubSubClientBuilder):
    """

    Author: Nicola Ricciardi
    """

    base_client: PubTopicSubClient = field(
        default_factory=lambda: PubTopicSubClient([], []),
        kw_only=True
    )

    @override
    def build(self) -> PubTopicSubClient:
        return self.base_client


from abc import abstractmethod
from collections.abc import Buffer
from typing import Self, Tuple
from dataclasses import dataclass, asdict
from busline.event.event import EventPayload
import pickle
import json

from busline.utils.serde import SerdableMixin

JSON_FORMAT_TYPE = "json"


@dataclass(frozen=True)
class JsonEventPayload(EventPayload, SerdableMixin):

    @classmethod
    @abstractmethod
    def from_json(cls, json_str: str) -> Self:
        raise NotImplemented()

    def serialize(self) -> Tuple[str, bytes]:
        return JSON_FORMAT_TYPE, pickle.dumps(json.dumps(asdict(self)))

    @classmethod
    def deserialize(cls, format_type: str, serialized_payload: bytes) -> Self:
        if format_type != JSON_FORMAT_TYPE:
            raise ValueError("Unsupported format type")

        return cls.from_json(pickle.loads(serialized_payload))



from abc import abstractmethod
from collections.abc import Buffer
from typing import Self, Tuple
from dataclasses import dataclass, asdict
from dataclasses_avroschema import AvroModel
from busline.event.event import EventPayload
from busline.utils.serde import SerdableMixin

AVRO_CONTENT_TYPE = "avro"


@dataclass(frozen=True)
class AvroEventPayload(EventPayload, SerdableMixin, AvroModel):


    def serialize(self) -> Tuple[str, bytes]:
        return AVRO_CONTENT_TYPE, AvroModel.serialize(self, serialization_type=AVRO_CONTENT_TYPE)

    @classmethod
    def deserialize(cls, format_type: str, serialized_payload: bytes) -> Self:
        if format_type != AVRO_CONTENT_TYPE:
            raise ValueError("Unsupported format type")

        return AvroModel.deserialize.__func__(cls, serialized_payload, serialization_type=AVRO_CONTENT_TYPE)
        # __func__ instead of __call__ because `AvroModel.deserialize` is a classmethod



import logging
from typing import override

from busline.client.publisher.publisher import Publisher
from busline.event.event import Event
from busline.local.eventbus.eventbus import EventBus
from busline.exceptions import EventBusClientNotConnected
from dataclasses import dataclass, field


@dataclass
class LocalEventBusPublisher(Publisher):
    """
    Publisher which works with local eventbus, this class can be initialized and used stand-alone

    Author: Nicola Ricciardi
    """

    eventbus: EventBus
    connected: bool = field(default=False)

    @override
    async def connect(self):
        logging.info(f"publisher {self.identifier} connecting...")
        self.connected = True

    @override
    async def disconnect(self):
        logging.info(f"publisher {self.identifier} disconnecting...")
        self.connected = False

    @override
    async def _internal_publish(self, topic_name: str, event: Event, **kwargs):

        if not self.connected:
            raise EventBusClientNotConnected()

        await self.eventbus.put_event(topic_name, event)


import unittest

from busline.client.multiclient import EventBusMultiClient
from busline.client.subscriber.event_handler.closure_event_handler import ClosureEventHandler
from busline.local.eventbus.async_local_eventbus import AsyncLocalEventBus
from busline.local.eventbus.local_eventbus import LocalEventBus
from busline.local.local_pubsub_client import LocalPubSubClientBuilder
from busline.local.publisher.local_publisher import LocalEventBusPublisher
from busline.event.event import Event
from busline.local.subscriber.local_subscriber import LocalEventBusSubscriber


class TestLocalEventBus(unittest.IsolatedAsyncioTestCase):

    async def test_async_eventbus(self):

        local_eventbus_instance1 = LocalEventBus()       # singleton
        local_eventbus_instance2 = LocalEventBus()       # singleton

        self.assertIs(local_eventbus_instance1, local_eventbus_instance2)        # check singleton

        event = Event()
        received_event = None

        def callback(t: str, e: Event):
            nonlocal received_event

            received_event = e

        subscriber = LocalEventBusSubscriber(
            eventbus=local_eventbus_instance1,
            fallback_event_handler=ClosureEventHandler(callback)
        )
        publisher = LocalEventBusPublisher(eventbus=local_eventbus_instance2)

        await subscriber.connect()
        await publisher.connect()

        await subscriber.subscribe("test")

        await publisher.publish("test", event)

        self.assertIs(event, received_event)

        await subscriber.unsubscribe()
        received_event = None

        await publisher.publish("test", event)

        self.assertIs(received_event, None)


    async def test_local_client(self):
        received_event = None
        event = Event()

        def client_callback(topic_name: str, e: Event):
            nonlocal received_event

            received_event = e

        client = LocalPubSubClientBuilder()\
                    .with_default_publisher()\
                    .with_closure_subscriber(client_callback)\
                    .build()

        await client.connect()

        await client.subscribe("test")

        await client.publish("test", event)

        self.assertIs(event, received_event)

        await client.unsubscribe()
        received_event = None

        await client.publish("test", event)

        self.assertIs(received_event, None)

    async def test_mhs(self):

        received_event = 0

        def callback(t: str, e: Event):
            nonlocal received_event

            received_event += 1

        subscriber = LocalEventBusSubscriber(
            fallback_event_handler=ClosureEventHandler(callback),
            eventbus=LocalEventBus()
        )

        await subscriber.connect()

        await subscriber.subscribe("t1")
        await subscriber.subscribe("t2", handler=ClosureEventHandler(callback))

        await subscriber.notify("t1", Event())

        self.assertEqual(received_event, 1)

        await subscriber.notify("t2", Event())

        self.assertEqual(received_event, 2)

        await subscriber.unsubscribe()

        await subscriber.notify("t1", Event())
        await subscriber.notify("t2", Event())

        self.assertEqual(received_event, 2)

    async def test_multi_client(self):
        local_eventbus_instance1 = AsyncLocalEventBus()  # not singleton
        local_eventbus_instance2 = AsyncLocalEventBus()  # not singleton

        n_events: int = 0

        def on_event_callback(topic_name: str, e: Event):
            nonlocal n_events

            n_events += 1

        client1 = LocalPubSubClientBuilder(local_eventbus_instance1)\
                    .with_default_publisher()\
                    .with_closure_subscriber(lambda t, e: ...)\
                    .build()

        client2 = LocalPubSubClientBuilder(local_eventbus_instance2)\
                    .with_default_publisher()\
                    .with_closure_subscriber(lambda t, e: ...)\
                    .build()

        multi_client = EventBusMultiClient([
            client1,
            client2
        ])

        await multi_client.connect()

        await multi_client.subscribe("topic", handler=ClosureEventHandler(on_event_callback))

        await multi_client.publish("topic", Event())

        await multi_client.disconnect()

        self.assertEqual(n_events, 2)




if __name__ == '__main__':
    unittest.main()


from busline.event.event import Event
from busline.local import DEFAULT_EVENT_BUS_INSTANCE
from busline.local.eventbus.eventbus import EventBus


class LocalEventBus(EventBus):
    """
    Local *singleton* event bus instance

    Author: Nicola Ricciardi
    """

    # === SINGLETON pattern ===
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = DEFAULT_EVENT_BUS_INSTANCE # super().__new__(cls)

        return cls._instance

    async def put_event(self, topic: str, event: Event):
        return self._instance.put_event(topic, event)


from __future__ import annotations

import uuid
import datetime
from dataclasses import dataclass, field
from typing import Optional
from abc import ABC
from collections.abc import Buffer
from busline.utils.serde import SerdableMixin


@dataclass(frozen=True)
class EventPayload(SerdableMixin, ABC):
    """
    Event payload for an event, it must be serializable and deserializable

    Author: Nicola Ricciardi
    """

    def into_event(self, /, **kwargs) -> Event:
        return Event(
            payload=self,
            **kwargs
        )

@dataclass(frozen=True, kw_only=True)
class Event:
    """
    Event publishable in an eventbus

    Author: Nicola Ricciardi
    """

    identifier: str = field(default_factory=lambda: str(uuid.uuid4()))
    payload: Optional[EventPayload] = field(default=None)
    event_type: Optional[str] = field(default=None)
    timestamp: float = field(default_factory=lambda: datetime.datetime.now(datetime.timezone.utc).timestamp())
    metadata: dict = field(default_factory=dict)

    def __post_init__(self):
        if self.event_type is None and self.payload is not None:
            object.__setattr__(self, 'event_type', self.payload.__class__.__name__)



from abc import ABC, abstractmethod
from typing import Tuple, Self


class SerializableMixin(ABC):
    """
    Author: Nicola Ricciardi
    """


    @abstractmethod
    def serialize(self) -> Tuple[str, bytes]:
        """
        Serialize itself and return the format as string
        """

        raise NotImplemented()


class DeserializableMixin(ABC):
    """
    Author: Nicola Ricciardi
    """

    @classmethod
    @abstractmethod
    def deserialize(cls, payload_type: str, payload: bytes) -> Self:
        raise NotImplemented()


class SerdableMixin(SerializableMixin, DeserializableMixin, ABC):
    pass

